//************************************************************************
//  DoubleDeque
//************************************************************************


#ifndef DOUBLEDEQUE_HH
#define DOUBLEDEQUE_HH

#include <vector>

template<typename ItemType>
class DoubleDeque
{
 public:
    class Row;

    ItemType& at(int rowIndex, int colIndex);

    const Row& getRow(int rowIndex) const;

    void clear();


    DoubleDeque() {}


//-----------------------------------------------------------------------
 private:
    template<typename T>
    class AutoDeque;

    AutoDeque<Row> rows;

    // Disable copying and assignment:
    DoubleDeque(const DoubleDeque&);
    const DoubleDeque& operator=(const DoubleDeque&);
};


template<typename ItemType>
class DoubleDeque<ItemType>::Row
{
 public:
    ItemType& at(int colIndex);
    const ItemType& at(int colIndex) const;


 private:
    AutoDeque<ItemType> data;
};



//=======================================================================
//=======================================================================
// Implementations
//=======================================================================
//=======================================================================

template<typename ItemType>
template<typename T>
class DoubleDeque<ItemType>::AutoDeque
{
    static const int BLOCK_SIZE = 65536;

    std::vector<T*> dataBlocks;
    int currentMinIndex, currentMaxIndex;
    int firstItemAbsoluteIndex;

    T defaultValue;

    int getBlockIndexOf(int index)
    {
        if(index >= 0) return index/BLOCK_SIZE;
        return (index-(BLOCK_SIZE-1))/BLOCK_SIZE;
    }


 public:
    AutoDeque():
        currentMinIndex(0), currentMaxIndex(-1), firstItemAbsoluteIndex(0)
    {}

    ~AutoDeque()
    {
        clear();
    }

    T& at(int index)
    {
        if(dataBlocks.empty())
        {
            dataBlocks.push_back(new T[BLOCK_SIZE]);
            currentMinIndex = index;
            currentMaxIndex = index;
            firstItemAbsoluteIndex =
                index - BLOCK_SIZE*getBlockIndexOf(index);
        }

        else if(index < currentMinIndex)
        {
            const int currentMinIndexBlock = getBlockIndexOf(currentMinIndex);
            const int newMinIndexBlock = getBlockIndexOf(index);
            if(newMinIndexBlock < currentMinIndexBlock)
            {
                const unsigned newBlocksAmount =
                    currentMinIndexBlock - newMinIndexBlock;

                // We assume this happens rarely so its speed doesn't
                // really matter:
                dataBlocks.insert(dataBlocks.begin(), newBlocksAmount, 0);

                for(unsigned i = 0; i < newBlocksAmount; ++i)
                {
                    dataBlocks[i] = new T[BLOCK_SIZE];
                }
            }

            currentMinIndex = index;
            firstItemAbsoluteIndex = index - BLOCK_SIZE*newMinIndexBlock;
        }

        else if(index > currentMaxIndex)
        {
            const int currentMaxIndexBlock = getBlockIndexOf(currentMaxIndex);
            const int newMaxIndexBlock = getBlockIndexOf(index);
            if(newMaxIndexBlock > currentMaxIndexBlock)
            {
                const unsigned newBlocksAmount =
                    newMaxIndexBlock - currentMaxIndexBlock;

                for(unsigned i = 0; i < newBlocksAmount; ++i)
                {
                    dataBlocks.push_back(new T[BLOCK_SIZE]);
                }
            }

            currentMaxIndex = index;
        }

        const int i = index - currentMinIndex + firstItemAbsoluteIndex;
        return dataBlocks[i/BLOCK_SIZE][i%BLOCK_SIZE];
    }

    const T& at(int index) const
    {
        if(index < currentMinIndex || index > currentMaxIndex)
            return defaultValue;

        const int i = index - currentMinIndex + firstItemAbsoluteIndex;
        return dataBlocks[i/BLOCK_SIZE][i%BLOCK_SIZE];
    }

    void clear()
    {
        currentMinIndex = 0; currentMaxIndex = -1;
        for(unsigned i = 0; i < dataBlocks.size(); ++i)
            delete[] dataBlocks[i];
        dataBlocks.clear();
    }


 private:
    // Disable copying and assignment:
    AutoDeque(const AutoDeque&);
    const AutoDeque& operator=(const AutoDeque&);
};


template<typename ItemType>
ItemType& DoubleDeque<ItemType>::Row::at(int colIndex)
{
    return data.at(colIndex);
}

template<typename ItemType>
const ItemType& DoubleDeque<ItemType>::Row::at(int colIndex) const
{
    return data.at(colIndex);
}

template<typename ItemType>
ItemType& DoubleDeque<ItemType>::at(int rowIndex, int colIndex)
{
    return rows.at(rowIndex).at(colIndex);
}

template<typename ItemType>
const typename DoubleDeque<ItemType>::Row&
DoubleDeque<ItemType>::getRow(int rowIndex) const
{
    return rows.at(rowIndex);
}

template<typename ItemType>
void DoubleDeque<ItemType>::clear()
{
    rows.clear();
}



#endif
